@@ -1,22 +1,23 @@
 /* SPDX-License-Identifier: LGPL-2.1-or-later */
 
+#include <stdlib.h>
+
 #include "alloc-util.h"
 #include "chase.h"
 #include "dirent-util.h"
 #include "env-file.h"
-#include "env-util.h"
+#include "errno-util.h"
 #include "fd-util.h"
-#include "fileio.h"
 #include "fs-util.h"
 #include "glyph-util.h"
-#include "macro.h"
+#include "log.h"
 #include "os-util.h"
-#include "parse-util.h"
 #include "path-util.h"
 #include "stat-util.h"
 #include "string-table.h"
 #include "string-util.h"
 #include "strv.h"
+#include "time-util.h"
 #include "utf8.h"
 #include "xattr-util.h"
 
@@ -121,7 +122,7 @@ static int extension_release_strict_xattr_value(int extension_release_fd, const
         assert(filename);
 
         /* No xattr or cannot parse it? Then skip this. */
-        r = getxattr_at_bool(extension_release_fd, /* path= */ NULL, "user.extension-release.strict", /* flags= */ 0);
+        r = getxattr_at_bool(extension_release_fd, /* path= */ NULL, "user.extension-release.strict", /* at_flags= */ 0);
         if (ERRNO_IS_NEG_XATTR_ABSENT(r))
                 return log_debug_errno(r, "%s/%s does not have user.extension-release.strict xattr, ignoring.",
                                        extension_release_dir_path, filename);
@@ -138,7 +139,7 @@ static int extension_release_strict_xattr_value(int extension_release_fd, const
 
         log_debug("%s/%s: 'user.extension-release.strict' attribute is false%s",
                   extension_release_dir_path, filename,
-                  special_glyph(SPECIAL_GLYPH_ELLIPSIS));
+                  glyph(GLYPH_ELLIPSIS));
 
         return false;
 }
